/*
 * Copyright (C) 2010-2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "Connection.h"
#include "MessageNames.h"
#include "SandboxExtension.h"
#include "SharedMemory.h"
#include "WebPasteboardProxyMessagesReplies.h"
#include <WebCore/Pasteboard.h>
#include <WebCore/PasteboardItemInfo.h>
#include <wtf/Forward.h>
#include <wtf/Optional.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace IPC {
class SharedBufferCopy;
}

namespace WebCore {
class Color;
class SelectionData;
struct PasteboardImage;
struct PasteboardURL;
struct PasteboardWebContent;
}

namespace Messages {
namespace WebPasteboardProxy {

static inline IPC::ReceiverName messageReceiverName()
{
    return IPC::ReceiverName::WebPasteboardProxy;
}

#if PLATFORM(IOS_FAMILY)
class WriteURLToPasteboard {
public:
    using Arguments = std::tuple<const WebCore::PasteboardURL&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_WriteURLToPasteboard; }
    static const bool isSync = false;

    WriteURLToPasteboard(const WebCore::PasteboardURL& url, const String& pasteboardName)
        : m_arguments(url, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class WriteWebContentToPasteboard {
public:
    using Arguments = std::tuple<const WebCore::PasteboardWebContent&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_WriteWebContentToPasteboard; }
    static const bool isSync = false;

    WriteWebContentToPasteboard(const WebCore::PasteboardWebContent& content, const String& pasteboardName)
        : m_arguments(content, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class WriteImageToPasteboard {
public:
    using Arguments = std::tuple<const WebCore::PasteboardImage&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_WriteImageToPasteboard; }
    static const bool isSync = false;

    WriteImageToPasteboard(const WebCore::PasteboardImage& pasteboardImage, const String& pasteboardName)
        : m_arguments(pasteboardImage, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class WriteStringToPasteboard {
public:
    using Arguments = std::tuple<const String&, const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_WriteStringToPasteboard; }
    static const bool isSync = false;

    WriteStringToPasteboard(const String& pasteboardType, const String& text, const String& pasteboardName)
        : m_arguments(pasteboardType, text, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class UpdateSupportedTypeIdentifiers {
public:
    using Arguments = std::tuple<const Vector<String>&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_UpdateSupportedTypeIdentifiers; }
    static const bool isSync = false;

    UpdateSupportedTypeIdentifiers(const Vector<String>& identifiers, const String& pasteboardName)
        : m_arguments(identifiers, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class WriteCustomData {
public:
    using Arguments = std::tuple<const Vector<WebCore::PasteboardCustomData>&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_WriteCustomData; }
    static const bool isSync = true;

    using DelayedReply = WriteCustomDataDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    WriteCustomData(const Vector<WebCore::PasteboardCustomData>& data, const String& pasteboardName)
        : m_arguments(data, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class TypesSafeForDOMToReadAndWrite {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_TypesSafeForDOMToReadAndWrite; }
    static const bool isSync = true;

    using DelayedReply = TypesSafeForDOMToReadAndWriteDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<String>& types);
    using Reply = std::tuple<Vector<String>&>;
    using ReplyArguments = std::tuple<Vector<String>>;
    TypesSafeForDOMToReadAndWrite(const String& pasteboardName, const String& origin)
        : m_arguments(pasteboardName, origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class AllPasteboardItemInfo {
public:
    using Arguments = std::tuple<const String&, int64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_AllPasteboardItemInfo; }
    static const bool isSync = true;

    using DelayedReply = AllPasteboardItemInfoDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<Vector<WebCore::PasteboardItemInfo>>& allInfo);
    using Reply = std::tuple<Optional<Vector<WebCore::PasteboardItemInfo>>&>;
    using ReplyArguments = std::tuple<Optional<Vector<WebCore::PasteboardItemInfo>>>;
    AllPasteboardItemInfo(const String& pasteboardName, int64_t changeCount)
        : m_arguments(pasteboardName, changeCount)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class InformationForItemAtIndex {
public:
    using Arguments = std::tuple<uint64_t, const String&, int64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_InformationForItemAtIndex; }
    static const bool isSync = true;

    using DelayedReply = InformationForItemAtIndexDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<WebCore::PasteboardItemInfo>& info);
    using Reply = std::tuple<Optional<WebCore::PasteboardItemInfo>&>;
    using ReplyArguments = std::tuple<Optional<WebCore::PasteboardItemInfo>>;
    InformationForItemAtIndex(uint64_t index, const String& pasteboardName, int64_t changeCount)
        : m_arguments(index, pasteboardName, changeCount)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetPasteboardItemsCount {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_GetPasteboardItemsCount; }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardItemsCountDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, uint64_t itemsCount);
    using Reply = std::tuple<uint64_t&>;
    using ReplyArguments = std::tuple<uint64_t>;
    explicit GetPasteboardItemsCount(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ReadStringFromPasteboard {
public:
    using Arguments = std::tuple<uint64_t, const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_ReadStringFromPasteboard; }
    static const bool isSync = true;

    using DelayedReply = ReadStringFromPasteboardDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& string);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    ReadStringFromPasteboard(uint64_t index, const String& pasteboardType, const String& pasteboardName)
        : m_arguments(index, pasteboardType, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ReadURLFromPasteboard {
public:
    using Arguments = std::tuple<uint64_t, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_ReadURLFromPasteboard; }
    static const bool isSync = true;

    using DelayedReply = ReadURLFromPasteboardDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& url, const String& title);
    using Reply = std::tuple<String&, String&>;
    using ReplyArguments = std::tuple<String, String>;
    ReadURLFromPasteboard(uint64_t index, const String& pasteboardName)
        : m_arguments(index, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ReadBufferFromPasteboard {
public:
    using Arguments = std::tuple<uint64_t, const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_ReadBufferFromPasteboard; }
    static const bool isSync = true;

    using DelayedReply = ReadBufferFromPasteboardDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::SharedMemory::IPCHandle& handle);
    using Reply = std::tuple<WebKit::SharedMemory::IPCHandle&>;
    using ReplyArguments = std::tuple<WebKit::SharedMemory::IPCHandle>;
    ReadBufferFromPasteboard(uint64_t index, const String& pasteboardType, const String& pasteboardName)
        : m_arguments(index, pasteboardType, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ContainsStringSafeForDOMToReadForType {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_ContainsStringSafeForDOMToReadForType; }
    static const bool isSync = true;

    using DelayedReply = ContainsStringSafeForDOMToReadForTypeDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool result);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    ContainsStringSafeForDOMToReadForType(const String& type, const String& pasteboardName)
        : m_arguments(type, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(COCOA)
class GetNumberOfFiles {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_GetNumberOfFiles; }
    static const bool isSync = true;

    using DelayedReply = GetNumberOfFilesDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, uint64_t numberOfFiles);
    using Reply = std::tuple<uint64_t&>;
    using ReplyArguments = std::tuple<uint64_t>;
    explicit GetNumberOfFiles(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardTypes {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_GetPasteboardTypes; }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardTypesDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<String>& types);
    using Reply = std::tuple<Vector<String>&>;
    using ReplyArguments = std::tuple<Vector<String>>;
    explicit GetPasteboardTypes(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardPathnamesForType {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_GetPasteboardPathnamesForType; }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardPathnamesForTypeDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<String>& pathnames, const WebKit::SandboxExtension::HandleArray& sandboxExtensions);
    using Reply = std::tuple<Vector<String>&, WebKit::SandboxExtension::HandleArray&>;
    using ReplyArguments = std::tuple<Vector<String>, WebKit::SandboxExtension::HandleArray>;
    GetPasteboardPathnamesForType(const String& pasteboardName, const String& pasteboardType)
        : m_arguments(pasteboardName, pasteboardType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardStringForType {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_GetPasteboardStringForType; }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardStringForTypeDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& string);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    GetPasteboardStringForType(const String& pasteboardName, const String& pasteboardType)
        : m_arguments(pasteboardName, pasteboardType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardStringsForType {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_GetPasteboardStringsForType; }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardStringsForTypeDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<String>& strings);
    using Reply = std::tuple<Vector<String>&>;
    using ReplyArguments = std::tuple<Vector<String>>;
    GetPasteboardStringsForType(const String& pasteboardName, const String& pasteboardType)
        : m_arguments(pasteboardName, pasteboardType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardBufferForType {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_GetPasteboardBufferForType; }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardBufferForTypeDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::SharedMemory::IPCHandle& handle);
    using Reply = std::tuple<WebKit::SharedMemory::IPCHandle&>;
    using ReplyArguments = std::tuple<WebKit::SharedMemory::IPCHandle>;
    GetPasteboardBufferForType(const String& pasteboardName, const String& pasteboardType)
        : m_arguments(pasteboardName, pasteboardType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardChangeCount {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_GetPasteboardChangeCount; }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardChangeCountDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    explicit GetPasteboardChangeCount(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardColor {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_GetPasteboardColor; }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardColorDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::Color& color);
    using Reply = std::tuple<WebCore::Color&>;
    using ReplyArguments = std::tuple<WebCore::Color>;
    explicit GetPasteboardColor(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardURL {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_GetPasteboardURL; }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardURLDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& urlString);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    explicit GetPasteboardURL(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class AddPasteboardTypes {
public:
    using Arguments = std::tuple<const String&, const Vector<String>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_AddPasteboardTypes; }
    static const bool isSync = true;

    using DelayedReply = AddPasteboardTypesDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    AddPasteboardTypes(const String& pasteboardName, const Vector<String>& pasteboardTypes)
        : m_arguments(pasteboardName, pasteboardTypes)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetPasteboardTypes {
public:
    using Arguments = std::tuple<const String&, const Vector<String>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_SetPasteboardTypes; }
    static const bool isSync = true;

    using DelayedReply = SetPasteboardTypesDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    SetPasteboardTypes(const String& pasteboardName, const Vector<String>& pasteboardTypes)
        : m_arguments(pasteboardName, pasteboardTypes)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetPasteboardURL {
public:
    using Arguments = std::tuple<const WebCore::PasteboardURL&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_SetPasteboardURL; }
    static const bool isSync = true;

    using DelayedReply = SetPasteboardURLDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    SetPasteboardURL(const WebCore::PasteboardURL& pasteboardURL, const String& pasteboardName)
        : m_arguments(pasteboardURL, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetPasteboardColor {
public:
    using Arguments = std::tuple<const String&, const WebCore::Color&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_SetPasteboardColor; }
    static const bool isSync = true;

    using DelayedReply = SetPasteboardColorDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    SetPasteboardColor(const String& pasteboardName, const WebCore::Color& color)
        : m_arguments(pasteboardName, color)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetPasteboardStringForType {
public:
    using Arguments = std::tuple<const String&, const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_SetPasteboardStringForType; }
    static const bool isSync = true;

    using DelayedReply = SetPasteboardStringForTypeDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    SetPasteboardStringForType(const String& pasteboardName, const String& pasteboardType, const String& string)
        : m_arguments(pasteboardName, pasteboardType, string)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetPasteboardBufferForType {
public:
    using Arguments = std::tuple<const String&, const String&, const WebKit::SharedMemory::IPCHandle&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_SetPasteboardBufferForType; }
    static const bool isSync = true;

    using DelayedReply = SetPasteboardBufferForTypeDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    SetPasteboardBufferForType(const String& pasteboardName, const String& pasteboardType, const WebKit::SharedMemory::IPCHandle& handle)
        : m_arguments(pasteboardName, pasteboardType, handle)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class ContainsURLStringSuitableForLoading {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_ContainsURLStringSuitableForLoading; }
    static const bool isSync = true;

    using DelayedReply = ContainsURLStringSuitableForLoadingDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool result);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit ContainsURLStringSuitableForLoading(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class URLStringSuitableForLoading {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_URLStringSuitableForLoading; }
    static const bool isSync = true;

    using DelayedReply = URLStringSuitableForLoadingDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& url, const String& title);
    using Reply = std::tuple<String&, String&>;
    using ReplyArguments = std::tuple<String, String>;
    explicit URLStringSuitableForLoading(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK)
class GetTypes {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_GetTypes; }
    static const bool isSync = true;

    using DelayedReply = GetTypesDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<String>& types);
    using Reply = std::tuple<Vector<String>&>;
    using ReplyArguments = std::tuple<Vector<String>>;
    explicit GetTypes(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK)
class ReadText {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_ReadText; }
    static const bool isSync = true;

    using DelayedReply = ReadTextDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& text);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    explicit ReadText(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK)
class ReadFilePaths {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_ReadFilePaths; }
    static const bool isSync = true;

    using DelayedReply = ReadFilePathsDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<String>& types);
    using Reply = std::tuple<Vector<String>&>;
    using ReplyArguments = std::tuple<Vector<String>>;
    explicit ReadFilePaths(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK)
class ReadBuffer {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_ReadBuffer; }
    static const bool isSync = true;

    using DelayedReply = ReadBufferDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const IPC::SharedBufferCopy& data);
    using Reply = std::tuple<IPC::SharedBufferCopy&>;
    using ReplyArguments = std::tuple<IPC::SharedBufferCopy>;
    ReadBuffer(const String& pasteboardName, const String& pasteboardType)
        : m_arguments(pasteboardName, pasteboardType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK)
class WriteToClipboard {
public:
    using Arguments = std::tuple<const String&, const WebCore::SelectionData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_WriteToClipboard; }
    static const bool isSync = false;

    WriteToClipboard(const String& pasteboardName, const WebCore::SelectionData& pasteboardContent)
        : m_arguments(pasteboardName, pasteboardContent)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK)
class ClearClipboard {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_ClearClipboard; }
    static const bool isSync = false;

    explicit ClearClipboard(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if USE(LIBWPE)
class GetPasteboardTypes {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_GetPasteboardTypes; }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardTypesDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<String>& types);
    using Reply = std::tuple<Vector<String>&>;
    using ReplyArguments = std::tuple<Vector<String>>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if USE(LIBWPE)
class WriteWebContentToPasteboard {
public:
    using Arguments = std::tuple<const WebCore::PasteboardWebContent&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_WriteWebContentToPasteboard; }
    static const bool isSync = false;

    explicit WriteWebContentToPasteboard(const WebCore::PasteboardWebContent& content)
        : m_arguments(content)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if USE(LIBWPE)
class WriteStringToPasteboard {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPasteboardProxy_WriteStringToPasteboard; }
    static const bool isSync = false;

    WriteStringToPasteboard(const String& pasteboardType, const String& text)
        : m_arguments(pasteboardType, text)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

} // namespace WebPasteboardProxy
} // namespace Messages
